import numpy as np
import matplotlib.pyplot as pl
from sympy import factorial, sympify, Rational
#from sage.combinat.combinat import permutations
from utils import permutations

#=====================================================
class RootedTree(str):
#=====================================================
    r"""
        A rooted tree is a directed acyclic graph with one node, which
        has no incoming edges, designated as the root.
        Rooted trees are useful for analyzing the order conditions of 
        multistage numerical ODE solvers, such as Runge-Kutta methods 
        and other general linear methods.

        The trees are represented as strings, using one of the notations
        introduced by Butcher (the third column of Table 300(I) of
        Butcher's text).  The character 'T' is used in place of $\\tau$ 
        to represent a vertex, and braces '{ }' are used instead of brackets 
        '[ ]' to indicate that everything inside the braces
        is joined to a single parent node.  Thus the first four trees are:

        'T', '{T}', '{T^2}', {{T}}'

        These can be generated using the function list_trees(), which
        returns a list of all trees of a given order::

            >>> from nodepy import *
            >>> for p in range(4): print rt.list_trees(p)
            ['']
            ['T']
            ['{T}']
            ['{{T}}', '{T^2}']
            
        Note that the tree of order 0 is indicated by an empty string.
        
        If the tree contains an edge from vertex A to vertex B, vertex
        B is said to be a child of vertex A.
        A vertex with no children is referred to as a leaf.  

        .. warning::

            One important convention is assumed in the code; namely, that at each 
            level, leaves are listed first (before any other subtrees), 
            and if there are $n$ leaves, we write 'T^n'.
        
        .. note::

            Currently, powers cannot be used for subtrees; thus

            '{{T}{T}}'

            is valid, while

            '{{T}^2}' 

            is not.  This restriction may be lifted in the future.

        **Examples**::

            >>> from nodepy import rooted_trees as rt
            >>> tree=rt.RootedTree('{T^2{T{T}}{T}}')
            >>> tree.order()
            9
            >>> tree.density()
            144
            >>> tree.symmetry()
            2

        Topologically equivalent trees are considered equal::

            >>> tree2=RootedTree('{T^2{T}{T{T}}}')
            >>> tree2==tree
            True

        We can generate Python code to evaluate the elementary weight
        corresponding to a given tree for a given class of methods::

            >>> rk.elementary_weight_str(tree)
            'dot(b,dot(A,c)*dot(A,c*dot(A,c))*c**2)'

        **References**:  
            #. [butcher2003]_
            #. [hairer1993]_
    """
    def __init__(self,strg):
        """
            TODO:   - Check validity of strg more extensively
                    - Accept any leaf ordering, but convert it to our convention
                    - convention for ordering of subtrees?
        """
        if any([strg[i] not in '{}T^1234567890' for i in range(len(strg))]):
            raise Exception('Not a valid rooted tree string (illegal character)')
        op,cl=strg.count('{'),strg.count('}')
        if op!=cl or (op+cl>0 and (strg[0]!='{' or strg[-1]!='}')):
            raise Exception('Not a valid rooted tree string')
        self=strg

    def order(self):
        """
        The order of a rooted tree, denoted $r(t)$, is the number of 
        vertices in the tree.

        **Examples**::

            >>> from nodepy import rooted_trees as rt
            >>> tree=rt.RootedTree('{T^2{T{T}}}')
            >>> tree.order()
            7
        """
        if self=='T': return 1
        if self=='':  return 0
        r=self.count('{')
        pos=0
        while pos!=-1:
            pos=self.find('T',pos+1)
            if pos!=-1:
                try: r+=int(self[pos+2])
                except: r+=1
        return r

    def density(self):
        r"""
        The density of a rooted tree, denoted by $\\gamma(t)$,
        is the product of the orders of the subtrees.

        **Examples**::

            >>> from nodepy import rooted_trees as rt
            >>> tree=rt.RootedTree('{T^2{T{T}}}')
            >>> tree.density()
            56

        **Reference**: 

            - [butcher2003]_ p. 127, eq. 301(c)
        """
        gamma=self.order()
        nleaves,subtrees=self._parse_subtrees()
        for tree in subtrees:
            gamma*=tree.density()
        return gamma

    def symmetry(self):
        r""" 
        The symmetry $\\sigma(t)$ of a rooted tree is...

        **Examples**::

            >>> from nodepy import rooted_trees as rt
            >>> tree=rt.RootedTree('{T^2{T{T}}}')
            >>> tree.symmetry()
            2

        **Reference**: 

            - [butcher2003]_ p. 127, eq. 301(b)
        """
        if self=='T': return 1
        sigma=1
        if self[1]=='T':
            try: sigma=factorial(int(self[3]))
            except: pass
        nleaves,subtrees=self._parse_subtrees()
        while len(subtrees)>0:
            st=subtrees[0]
            nst=subtrees.count(st)
            sigma*=factorial(nst)*st.symmetry()**nst
            while st in subtrees: subtrees.remove(st)
        return sigma


    def Dmap(self):
        """ 
        Butcher's function $D(t)$.  Represents differentiation.
        Defined by $D(t)$=0 except for D('T')=1.

        **Reference**: 
            #. [butcher1997]_
        """
        return self=='T'

    def lamda(self,alpha,extraargs=[]):
        r""" 
            Computes Butcher's functional lambda on a given tree
            for the function alpha.  This is used to compute the
            product of two functions on trees.

            *INPUT*:

                * alpha -- a function on rooted trees
                * extraargs -- a list containing any additional arguments 
                  that must be passed to alpha

            *OUTPUT*:
                * tprod -- a list of trees [t1, t2, ...]
                * fprod -- a list of numbers [a1, a2, ...]

            The meaning of the output is that 
            $\\lambda(\\alpha,t)(\\beta)=a1*\\beta(t1)+a2*\\beta(t2)+...$ 

            **Examples**::

                >>> from nodepy import rt
                >>> tree = rt.RootedTree('{T{T}}')
                >>> tree.lamda(rt.Emap)
                (['T', '{T}', '{{T}}', '{T}', '{T^2}', '{T{T}}'], [1/2, 1, 1, 1/2, 1, 1])

            **Reference**: 
                [butcher2003]_ pp. 275-276
        """
        if self=='': return [RootedTree('')],[0]
        if self=='T': return [RootedTree('T')],[1]
        t,u=self._factor()
        if extraargs:
            l1,f1=t.lamda(alpha,*extraargs)
            l2,f2=u.lamda(alpha,*extraargs)
            alphau=alpha(u,*extraargs)
        else:
            l1,f1=t.lamda(alpha)
            l2,f2=u.lamda(alpha)
            alphau=alpha(u)
        tprod=l1
        fprod=[alphau*f1i for f1i in f1 if f1!=0]
        #FOIL:
        for i in range(len(l1)):
            if f1!=0:
                for j in range(len(l2)):
                    if f2!=0:
                        tprod.append(l1[i]*l2[j])
                        fprod.append(f1[i]*f2[j])
        return tprod,fprod

    def lamda_str(self,alpha,extraargs=[]):
        """
            Alternate version of lamda above, but returns a string.
            Hopefully we can get rid of this (and the other str functions)
            when SAGE can handle noncommutative symbolic algebra.
        """
        if not isinstance(extraargs,list): extraargs=[extraargs]
        if self=='': return [RootedTree('')],[0]
        if self=='T': return [RootedTree('T')],[1]
        t,u=self._factor()
        if extraargs:
            l1,f1=t.lamda_str(alpha,*extraargs)
            l2,f2=u.lamda_str(alpha,*extraargs)
            alphau=alpha(u,*extraargs)
        else:
            l1,f1=t.lamda_str(alpha)
            l2,f2=u.lamda_str(alpha)
            alphau=alpha(u)
        tprod=l1
        fprod=[str(f1i)+'*'+alphau for f1i in f1 if f1!=0]
        #FOIL:
        for i in range(len(l1)):
            if f1!=0:
                for j in range(len(l2)):
                    if f2!=0:
                        tprod.append(l1[i]*l2[j])
                        fprod.append(str(f1[i])+'*'+str(f2[j]))
        return tprod,fprod

    def _factor(self):
        """ 
            Returns two rooted trees, t and u, such that self=t*u.

            **Input**: 
                - self -- any rooted tree
            **Output**: 
                - t,u -- a pair of rooted trees whose product t*u is equal to self.

            **Examples**::

                >>> tree=RootedTree('{T^2{T}}')
                >>> t,u=tree._factor()
                >>> t
                '{T{T}}'
                >>> u
                'T'
                >>> t*u==tree
                True

            .. note:: This function is typically only called by lamda().
        """
        nleaves,subtrees=self._parse_subtrees()
        if nleaves==0: # Root has no leaves
            t=RootedTree('{'+''.join(subtrees[1:])+'}')
            u=RootedTree(subtrees[0])
        if nleaves==1:
            t=RootedTree(self[0]+self[2:])
            u=RootedTree('T')
        if nleaves==2:
            t=RootedTree(self[0:2]+self[4:])
            u=RootedTree('T')
        if nleaves>2:
            t=RootedTree(self[0:3]+str(int(self[3])-1)+self[4:])
            u=RootedTree('T')
        if t=='{}': t=RootedTree('T')
        return t,u

    def Gprod(self,alpha,beta,alphaargs=[],betaargs=[]):
        r""" 
            Returns the product of two functions on a given tree.

            INPUT:
                alpha, beta -- two functions on rooted trees
                               that return symbolic or numeric values
                alphaargs   -- a string containing any additional arguments 
                                that must be passed to function alpha
                betaargs    -- a string containing any additional arguments 
                                that must be passed to function beta

            OUTPUT:
                (alpha*beta)(self)
                i.e., the function that is the product (in G) of the
                functions alpha and beta.  Note that this product is
                not commutative.

            The product is given by

            $(\\alpha*\\beta)('')=\\beta('')$

            $(\\alpha*\\beta)(t) = \\lambda(\\alpha,t)(\\beta) + \\alpha(t)\\beta('')$

            .. note::
                Gprod can be used to compute products of more than two
                functions by passing Gprod itself in as beta, and providing
                the remaining functions to be multiplied as betaargs.

            **Examples**::

                >>> from nodepy import rt
                >>> tree = rt.RootedTree('{T{T}}')
                >>> tree.Gprod(rt.Emap,Dmap)
                1/2

            **Reference**: [butcher2003]_ p. 276, Thm. 386A 
        """
        trees,factors=self.lamda(alpha,*alphaargs)
        s=0
        for i in range(len(trees)):
            s+=factors[i]*beta(trees[i],*betaargs)
        s+=alpha(self,*alphaargs)*beta(RootedTree(""),*betaargs)
        return s

    def Gprod_str(self,alpha,beta,alphaargs=[],betaargs=[]):
        """ 
            Alternate version of Gprod, but operates on strings.
            Hopefully can be eliminated later in favor of symbolic
            manipulation.
        """
        trees,facs=self.lamda_str(alpha,*alphaargs)
        s=""
        for i in range(len(trees)):
            if facs[i]!=0:
                bet=beta(trees[i],*betaargs)
                if bet not in ['0','']:
                    if i>0: s+='+'
                    s+=str(facs[i])+"*"+bet
        bet=beta(RootedTree(""),*betaargs)
        if bet not in ['0','']:
            alph=alpha(self,*alphaargs)
            s+="+"+str(sympify(alph+'*'+bet))
        return s

    def plot(self,nrows=1,ncols=1,iplot=1,ttitle=''):
        """
            Plots the rooted tree.

            *INPUT*: (optional)
                * nrows, ncols -- number of rows and columns of subplots
                  in the figure
                * iplot        -- index of the subplot in which to plot
                  this tree

            These are only necessary if plotting more than one tree
            in a single figure using subplot.

            *OUTPUT*: None.

            The plot is created recursively by
            plotting the root, parsing the subtrees, plotting the 
            subtrees' roots, and calling _plot_subtree on each child
        """
        if iplot==1: pl.clf()
        pl.subplot(nrows,ncols,iplot)
        pl.hold(True)
        pl.scatter([0],[0])
        if self!='T': self._plot_subtree(0,0,1.)

        fs=int(np.ceil(30./nrows))
        pl.title(ttitle,{'fontsize': fs})
        pl.xticks([])
        pl.yticks([])
        pl.hold(False)
        pl.axis('off')
        #pl.show()
        #pl.ioff()


    def _plot_subtree(self,xroot,yroot,xwidth):
        """
            Recursively plots subtrees.  Should only be called from plot().

            INPUT:
                xroot, yroot -- coordinates at which root of this subtree 
                                is plotted
                xwidth -- width in which this subtree must fit, in order
                            to avoid possibly overlapping with others
        """
        ychild=yroot+1
        nleaves,subtrees=self._parse_subtrees()
        nchildren=nleaves+len(subtrees)

        dist=xwidth*(nchildren-1)/2.
        xchild=np.linspace(xroot-dist,xroot+dist,nchildren)
        pl.scatter(xchild,ychild*np.ones(nchildren))
        for i in range(nchildren):
            pl.plot([xroot,xchild[i]],[yroot,ychild],'-k')
            if i>nleaves-1:
                subtrees[i-nleaves]._plot_subtree(xchild[i],ychild,xwidth/3.)


    def _parse_subtrees(self):
        """ 
            Returns the number of leaves and a list of the subtrees,
            for a given rooted tree.

            OUTPUT:
                nleaves  -- number of leaves attached directly to the root
                subtrees -- list of non-leaf subtrees attached to the root

            The method can be thought of as returning what remains if the
            root of the tree is removed.  For efficiency, instead of
            returning possibly many copies of 'T', the leaves are just
            returned as a number.
        """
        from strmanip import get_substring, open_to_close
        if str(self)=='T' or str(self)=='': return 0,[]
        pos=0
        #Count leaves at current level
        if self[1]=='T':    
            if self[2]=='^': nleaves=int(self[3])
            else: nleaves=1
        else: nleaves=0

        subtrees=[]
        while pos!=-1:
            pos=self.find('{',pos+1)
            if pos!=-1:
                subtrees.append(RootedTree(get_substring(self,pos)))
                pos=open_to_close(self,pos)

        return nleaves,subtrees

    def list_equivalent_trees(self):
        """ 
            Returns a list of all strings (subject to our assumptions)
            equivalent to a given tree 

            INPUT: 
                self     -- any rooted tree
            OUTPUT:  
                treelist -- a list of all the 'legal' tree strings
                            that produce the same tree.

            The list of equivalent trees is obtained by taking all
            permutations of the (non-leaf) subtrees.
            This routine is used to test equality of trees.
        """
        nleaves,subtrees=self._parse_subtrees()
        if len(subtrees)==0: return [self]
        for i in range(len(subtrees)): subtrees[i]=str(subtrees[i])
        treelist = [RootedTree('{'+_powerString('T',nleaves,powchar='^')+
                    ''.join(sts)+'}') for sts in permutations(subtrees)]
        return treelist

    def __eq__(self,tree2):
        """
            Test equivalence of two rooted trees.
            Generates all 'legal' strings equivalent to the first
            tree, and checks whether the second is in that list.
        """
        ts=[str(t) for t in self.list_equivalent_trees()]
        if str(tree2) in ts: return True
        else: return False

    def __mul__(self,tree2):
        """ 
            Returns Butcher's product: t*u is the tree obtained by
            attaching the root of u as a child to the root of t. 
        """
        if self=='T': return RootedTree('{'+tree2+'}')
        if tree2=='T':  # We're just adding a leaf to self
            nleaves,subtrees=self._parse_subtrees()
            if nleaves==0: return RootedTree(self[0]+'T'+self[1:])
            if nleaves==1: return RootedTree(self[0]+'T^2'+self[2:])
            if nleaves>1:
                return RootedTree(self[0:3]+str(int(self[3])+1)+self[4:])
        else: return RootedTree(self[:-1]+tree2+'}') # tree2 wasn't just 'T'
#=====================================================
#End of RootedTree class
#=====================================================

#=====================================================
def plot_all_trees(p,title='str'):
#=====================================================
    """ Plots all rooted trees of order p """
    forest=list_trees(p)
    nplots=len(forest)
    nrows=int(np.ceil(np.sqrt(float(nplots))))
    ncols=int(np.floor(np.sqrt(float(nplots))))
    if nrows*ncols<nplots: ncols=ncols+1
    for tree in forest:
        if title=='str': ttitle=tree
        else: ttitle=''
        tree.plot(nrows,ncols,forest.index(tree)+1,ttitle=ttitle)
    fig=pl.figure(1)
    pl.setp(fig,facecolor='white')

#=====================================================
def list_trees(p,ind='all'):
#=====================================================
    """ 
    Returns rooted trees of order p.

    INPUT: 
    
        - p   -- order of trees desired
        - ind -- if given, returns a single tree corresponding to this index.
                Not very useful since the ordering isn't obvious.
                
    OUTPUT: list of all trees of order p (or just one, if ind is provided).
    
    Generates the rooted trees using Albrecht's 'Recursion 3'.

    **Examples**:

    Produce column of Butcher's Table 302(I)::

        >>> for i in range(1,11): 
        ...     forest=list_trees(i)
        ...     print len(forest)
        1
        1
        2
        4
        9
        20
        48
        115
        286
        719

    .. warning::

        This code is complete only up to order 10.  We need to extend it 
        by adding more subloops for p>10.

    TODO: Implement Butcher's formula (Theorem 302B) for the number
            of trees and determine to what order this is valid.

    **Reference**: [albrecht1996]_
    """

    if p==0: return [RootedTree('')]
    W=[[],[]] #This way indices agree with Albrecht
    R=[[],[]]
    R.append([RootedTree("{T}")])
    W.append([RootedTree("{{T}}")])
    for i in range(3,p):
        #Construct R[i]
        ps=_powerString("T",i-1,powchar="^")
        R.append([RootedTree("{"+ps+"}")])
        for w in W[i-1]:
            R[i].append(w)
        #Construct W[i]
        #l=0:
        W.append([RootedTree("{"+R[i][0]+"}")])
        for r in R[i][1:]:
            W[i].append(RootedTree("{"+r+"}"))
        for l in range(1,i-1): #level 1
            for r in R[i-l]:
                ps=_powerString("T",l,powchar="^")
                W[i].append(RootedTree("{"+ps+r+"}"))
        for l in range(0,i-3): #level 2
            for n in range(2,i-l-1):
                m=i-n-l
                if m<=n: #Avoid duplicate conditions
                    for Rm in R[m]:
                        lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                        for Rn in R[n][lowlim:]:
                            ps=_powerString("T",l,powchar="^")
                            W[i].append(RootedTree("{"+ps+Rm+Rn+"}"))
        for l in range(0,i-5): #level 3
            for n in range(2,i-l-3):
                for m in range(2,i-l-n-1):
                    s=i-m-n-l
                    if m<=n and n<=s: #Avoid duplicate conditions
                        for Rm in R[m]:
                            lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                            for Rn in R[n][lowlim:]:
                                lowlim2=(n<s and [0] or [R[n].index(Rn)])[0]
                                for Rs in R[s][lowlim2:]:
                                    ps=_powerString("T",l,powchar="^")
                                    W[i].append(RootedTree("{"+ps+Rm+Rn+Rs+"}"))
        for l in range(0,i-7): #level 4
            for n in range(2,i-l-5):
                for m in range(2,i-l-n-3):
                    for s in range(2,i-l-n-m-1):
                        t=i-s-m-n-l
                        if s<=t and n<=s and m<=n: #Avoid duplicate conditions
                            for Rm in R[m]:
                                lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                                for Rn in R[n][lowlim:]:
                                    lowlim2=(n<s and [0] or [R[n].index(Rn)])[0]
                                    for Rs in R[s][lowlim2:]:
                                        lowlim3=(s<t and [0] or [R[s].index(Rs)])[0]
                                        for Rt in R[t]:
                                            ps=_powerString("T",l,powchar="^")
                                            W[i].append(RootedTree("{"+ps+Rm+Rn+Rs+Rt+"}"))
    # The recursion above generates all trees except the 'blooms'
    # Now add the blooms:
    W[0].append(RootedTree("T"))
    for i in range(1,p):
        ps=_powerString("T",i,powchar="^")
        W[i].append(RootedTree("{"+ps+"}"))
  
    if ind=='all': return W[p-1]
    else: return W[p-1][ind]
    

def _powerString(s,npow,powchar="**",trailchar=''):
    r"""Raise string s to power npow with additional formatting."""
    if npow==0:
        return ""
    else:
        if npow==1:
            return s+trailchar
        else:
            return s+powchar+str(npow)+trailchar


#=====================================================
# Functions on trees
#=====================================================

def Dprod(tree,alpha):
    """
    Evaluate (alpha*D)(t).  Note that this is not equal to (D*alpha)(t).
    This function is necessary (rather than just using Gprod)
    in order to avoid infinite recursions.

    **Examples**::

        >>> from nodepy import rt
        >>> tree = rt.RootedTree('{T{T}}')
        >>> Dprod(tree,Emap)
        1/2
    """
    if tree=='': return 0
    if tree=='T': return alpha(RootedTree(''))
    nleaves,subtrees=tree._parse_subtrees()
    result=alpha(RootedTree('T'))**nleaves
    for subtree in subtrees:
        result*=alpha(subtree)
    return result

def Dprod_str(tree,alpha):
    if tree=='': return '0'
    if tree=='T': return alpha(RootedTree(''))
    nleaves,subtrees=tree._parse_subtrees()
    result=_powerString(alpha(RootedTree('T')),nleaves)
    for subtree in subtrees:
        if result!='': result+='*'
        result+=alpha(subtree)
    return result

def Dmap(tree):
    """ 
    Butcher's function D(t).  Represents differentiation.
    Defined by D(t)=0 except for D('T')=1.
    """
    return 1*(tree=='T')

def Dmap_str(tree):
    return str(int(tree=='T'))

def Gprod(tree,alpha,beta,alphaargs='',betaargs=[]):
    """ Returns the product of two functions on a given tree.
         See Butcher p. 276, Thm. 386A """
    return tree.Gprod(alpha,beta,alphaargs,betaargs)

def Gprod_str(tree,alpha,beta,alphaargs='',betaargs=[]):
    return tree.Gprod_str(alpha,beta,alphaargs,betaargs)

def Emap(tree,a=1):
    """ 
    Butcher's function E^a(t).
    Gives the B-series for the exact solution advanced 'a' steps
    in time.

    **Examples**::

        >>> from nodepy import rooted_trees as rt
        >>> tree=rt.RootedTree('{T^2{T{T}}}')
        >>> rt.Emap(tree)
        1/56
        >>> rt.Emap(tree,a=2)
        16/7

    **Reference**: 

        [butcher1997]_
    """
    return Rational(a**tree.order(),(tree.density()))

def Emap_str(tree,a=1):
    return str(Rational(a**tree.order(),(tree.density())))


#=====================================================
def recursiveVectors(p,ind='all'):
#=====================================================
  """ 
    Generate recursive vectors using Albrecht's 'recursion 1'.
    These are essentially the order conditions for Runge-Kutta
    methods, excluding those that correspond to bushy trees.
    More specifically, these are the vectors that must be
    orthogonal to the vector of weights b.

    Note that the individual order conditions obtained from
    this algorithm are different from those obtained using Butcher's
    approach.  But as a set of conditions up to some order they
    are, of course, equivalent.

    Follows [albrecht1996]_ p. 1718

    .. warning::

        This code is complete only up to order 10.  We need to extend it 
        by adding more subloops for p>10.
  """
  if p>10: print 'recursiveVectors is not complete for orders p>10.'
  W=[[],[]]
  R=[[],[]]
  R.append(["tau[2]"])
  W.append(["tau[2]"])
  for i in range(3,p):
     #Construct R[i]
     R.append(["tau["+str(i)+"]"])
     for j in range(len(W[i-1])):
        R[i].append("A,"+W[i-1][j])
     #Construct W[i]
     #l=0:
     W.append(R[i][:])
     for l in range(1,i-1): #level 1
        ps=_powerString("C",l,trailchar=",")
        for j in range(len(R[i-l])):
          W[i].append(ps+R[i-l][j])
     for l in range(0,i-3): #level 2
        ps=_powerString("C",l,trailchar=",")
        for n in range(2,i-l-1):
          m=i-n-l
          if m<=n: #Avoid duplicate conditions
             for Rm in R[m]:
                lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                for Rn in R[n][lowlim:]:
                  W[i].append(ps+Rm+"*"+Rn)
     for l in range(0,i-5): #level 3
        ps=_powerString("C",l,trailchar=",")
        for n in range(2,i-l-3):
          for m in range(2,i-l-n-1):
             s=i-m-n-l
             if m<=n and n<=s: #Avoid duplicate conditions
                for Rm in R[m]:
                  lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                  for Rn in R[n][lowlim:]:
                     lowlim2=(n<s and [0] or [R[n].index(Rn)])[0]
                     for Rs in R[s][lowlim2:]:
                        W[i].append(ps+Rm+"*"+Rn+"*"+Rs)
     for l in range(0,i-7): #level 4
        ps=_powerString("C",l,trailchar=",")
        for n in range(2,i-l-5):
          for m in range(2,i-l-n-3):
             for s in range(2,i-l-n-m-1):
                t=i-s-m-n-l
                if s<=t and n<=s and m<=n: #Avoid duplicate conditions
                  for Rm in R[m]:
                     lowlim=(m<n and [0] or [R[m].index(Rm)])[0]
                     for Rn in R[n][lowlim:]:
                        lowlim2=(n<s and [0] or [R[n].index(Rn)])[0]
                        for Rs in R[s][lowlim2:]:
                          lowlim3=(s<t and [0] or [R[s].index(Rs)])[0]
                          for Rt in R[t]:
                             W[i].append(ps+Rm+"*"+Rn+"*"+Rs+"*"+Rt)

  if ind=='all': return W[p-1]
  else: return W[p-1][ind]


#=====================================================
#=====================================================
#DEPRECATED FUNCTIONS
#=====================================================
#=====================================================

def py2tex(codestr):
    """Convert a python code string to LaTex"""

    strout=codestr.replace("'","^T")
    strout=strout.replace("*","")
    strout=strout.replace(".^","^")
    strout='$'+strout+'$'
    return strout


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
**Examples**::

    >>> import linear_multistep_method as lm
    >>> ab3=lm.Adams_Bashforth(3)
    >>> ab3.order()
    3
    >>> bdf2=lm.backward_difference_formula(2)
    >>> bdf2.order()
    2
    >>> bdf2.is_zero_stable()
    True
    >>> bdf7=lm.backward_difference_formula(7)
    >>> bdf7.is_zero_stable()
    False
    >>> bdf3=lm.backward_difference_formula(3)
    >>> bdf3.A_alpha_stability()
    172
    >>> ssp32=lm.elm_ssp2(3)
    >>> ssp32.order()
    2
    >>> ssp32.ssp_coefficient()
    1/2

"""
from general_linear_method import GeneralLinearMethod
import numpy as np
import snp

#=====================================================
class LinearMultistepMethod(GeneralLinearMethod):
#=====================================================
    r"""
        Implementation of linear multistep methods in the form:

        `\alpha_k y_{n+k} + \alpha_{k-1} y_{n+k-1} + ... + \alpha_0 y_n
        = h ( \beta_k f_{n+k} + ... + \beta_0 f_n )`

        Methods are automatically normalized so that \alpha_k=1.

        Notes: Representation follows Hairer & Wanner p. 368, NOT Butcher.

        **References**:  
            #. [hairer1993]_ Chapter III
            #. [butcher2003]_
    """
    def __init__(self,alpha,beta,name='Linear multistep method'):
        self.beta=beta/(alpha[-1])
        self.alpha=alpha/(alpha[-1])
        self.name=name

    def __num__(self):
        """
        Returns a copy of the method but with floating-point coefficients.
        This is useful whenever we need to operate numerically without
        worrying about the representation of the method.
        """
        import copy
        numself = copy.deepcopy(self)
        if self.alpha.dtype==object:
            numself.alpha=np.array(self.alpha,dtype=np.float64)
            numself.beta=np.array(self.beta,dtype=np.float64)
        return numself

    def characteristic_polynomials(self):
        r"""
        Returns the characteristic polynomials (also known as generating
        polynomials) of a linear multistep method.  They are:

        `\rho(z) = \sum_{j=0}^k \alpha_k z^k`

        `\sigma(z) = \sum_{j=0}^k \beta_k z^k`

        **Examples**::
            
            >>> from nodepy import lm
            >>> ab5 = lm.Adams_Bashforth(5)
            >>> rho,sigma = ab5.characteristic_polynomials()
            >>> print rho
               5     4
            1 x - 1 x

            >>> print sigma
                  4         3         2
            2.64 x - 3.853 x + 3.633 x - 1.769 x + 0.3486

        **References**:
            #. [hairer1993]_ p. 370, eq. 2.4

        """
        rho=np.poly1d(self.alpha[::-1])
        sigma=np.poly1d(self.beta[::-1])
        return rho, sigma

    def order(self,tol=1.e-10):
        r""" Return the order of the local truncation error of a linear multistep method.
        
        **Examples**::

            >>> from nodepy import lm
            >>> am3=lm.Adams_Moulton(3)
            >>> am3.order()
            4
        """
        p = 0
        while True:
            if self._satisfies_order_conditions(p+1,tol):
                p = p + 1
            else:
                return p

    def _satisfies_order_conditions(self,p,tol):
        """ Return True if the linear multistep method satisfies 
            the conditions of order p (only) """
        ii=snp.arange(len(self.alpha))
        return abs(sum(ii**p*self.alpha-p*self.beta*ii**(p-1)))<tol

    def ssp_coefficient(self):
        r""" Return the SSP coefficient of the method.

         The SSP coefficient is given by
        
        `\min_{0 \le j < k} -\alpha_k/beta_k`

        if `\alpha_j<0` and `\beta_j>0` for all `j`, and is equal to
        zero otherwise.


        **Examples**::

            >>> from nodepy import lm
            >>> ssp32=lm.elm_ssp2(3)
            >>> ssp32.ssp_coefficient()
            1/2

            >>> bdf2=lm.backward_difference_formula(2)
            >>> bdf2.ssp_coefficient()
            0
        """
        if np.any(self.alpha[:-1]>0) or np.any(self.beta<0): 
            return 0

        return min([-self.alpha[j]/self.beta[j] 
                    for j in range(len(self.alpha)-1) if self.beta[j]!=0])


    def plot_stability_region(self,N=100,N2=1000,color='r',filled=True, alpha=1.):
        r""" 
            The region of absolute stability of a linear multistep method is
            the set

            `\{ z \in C : \rho(\zeta) - z \sigma(zeta) \text{ satisfies the root condition} \}`

            where `\rho(zeta)` and `\sigma(zeta)` are the characteristic
            functions of the method.

            Also plots the boundary locus, which is
            given by the set of points z:

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            Here `\rho` and `\sigma` are the characteristic polynomials 
            of the method.

            References:
                [leveque2007]_ section 7.6.1


            **Input**: (all optional)
                - N       -- Number of gridpoints to use in each direction
                - bounds  -- limits of plotting region
                - color   -- color to use for this plot
                - filled  -- if true, stability region is filled in (solid); otherwise it is outlined
        """
        import matplotlib.pyplot as plt
        from utils import find_plot_bounds
        numself = self.__num__()
        rho, sigma = self.__num__().characteristic_polynomials()
        mag = lambda z : _root_condition(rho-z*sigma)
        vmag = np.vectorize(mag)
        bounds = find_plot_bounds(vmag,guess=(-10,1,-5,5),N=101)

        y=np.linspace(bounds[2],bounds[3],N)
        Y=np.tile(y[:,np.newaxis],(1,N))
        x=np.linspace(bounds[0],bounds[1],N)
        X=np.tile(x,(N,1))
        Z=X+Y*1j

        R=1.5-vmag(Z)

        z = self._boundary_locus()

        if filled:
            plt.contourf(X,Y,R,[0,1],colors=color,alpha=alpha)
        else:
            plt.contour(X,Y,R,[0,1],colors=color,alpha=alpha)
        plt.title('Absolute Stability Region for '+self.name)
        plt.hold(True)
        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.plot(np.real(z),np.imag(z),color='k',linewidth=3)
        plt.axis(bounds)
        plt.hold(False)
        plt.draw()

    def plot_boundary_locus(self,N=1000):
        r"""Plot the boundary locus, which is
            given by the set of points

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            where `\rho` and `\sigma` are the characteristic polynomials 
            of the method.

            References:
                [leveque2007]_ section 7.6.1
        """
        import matplotlib.pyplot as plt

        z = self._boundary_locus()

        plt.figure()
        plt.plot(np.real(z),np.imag(z),color='k',linewidth=3)
        plt.axis('image')
        plt.hold(True)
        bounds = plt.axis()
        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.title('Boundary locus for '+self.name)
        plt.hold(False)
        plt.draw()


    def _boundary_locus(self, N=1000):
        r"""Compute the boundary locus, which is
            given by the set of points

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            where `\rho` and `\sigma` are the characteristic polynomials 
            of the method.

            References:
                [leveque2007]_ section 7.6.1
        """
        theta=np.linspace(0.,2*np.pi,N)
        zeta = np.exp(theta*1j)
        rho,sigma=self.__num__().characteristic_polynomials()
        z = rho(zeta)/sigma(zeta)

        return z

    def A_alpha_stability(self, N=1000, tol=1.e-14):
        r"""Angle of `A(\alpha)`-stability.
        
        The result is given in degrees.  The result is only accurate to
        about 1 degree, so we round down.
        
        **Examples**:

            >>> from nodepy import lm
            >>> bdf5 = lm.backward_difference_formula(5)
            >>> bdf5.A_alpha_stability()
            103
        """
        from math import atan2, floor

        z = self._boundary_locus(N)
        rad = map(atan2,np.imag(z),np.real(z))
        rad = np.mod(np.array(rad),2*np.pi)

        return int(floor(np.min(np.abs(np.where(np.real(z)<-tol,rad,1.e99)-np.pi))/np.pi*360))

    def is_explicit(self):
        return self.beta[-1]==0

    def is_zero_stable(self,tol=1.e-13):
        r""" True if the method is zero-stable.

        **Examples**::

            >>> from nodepy import lm
            >>> bdf5=lm.backward_difference_formula(5)
            >>> bdf5.is_zero_stable()
            True
        """
        rho, sigma = self.characteristic_polynomials()
        return _root_condition(rho,tol)

    def __len__(self):
        return len(self.alpha)
#======================================================

def _root_condition(p,tol=1.e-13):
    r""" True if the polynomial `p` has all roots inside
    the unit circle and roots on the boundary of the unit circle
    are simple.

    **Examples**::

        >>> from nodepy import lm
        >>> p = np.poly1d((1,2.5,2,0.5))
        >>> lm._root_condition(p)
        False
    """
    if max(np.abs(p.r))>(1+tol):
        return False

    mod_one_roots = [r for r in p.r if abs(abs(r)-1)<tol]
    for i,r1 in enumerate(mod_one_roots):
        for r2 in mod_one_roots[i+1:]:
            if abs(r1-r2)<tol:
                return False
    return True


#======================================================
# Families of multistep methods
#======================================================

def Adams_Bashforth(k):
    r""" 
    Construct the k-step, Adams-Bashforth method.
    The methods are explicit and have order k.
    They have the form:

    `y_{n+1} = y_n + h \sum_{j=0}^{k-1} \beta_j f(y_n-k+j+1)`

    They are generated using equations (1.5) and (1.7) from 
    [hairer1993]_ III.1, along with the binomial expansion.

    **Examples**::

        >>> import linear_multistep_method as lm
        >>> ab3=lm.Adams_Bashforth(3)
        >>> ab3.order()
        3

        References:
            #. [hairer1993]_
    """
    import sympy
    from sympy import Rational

    one = Rational(1,1)

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[k]=one
    alpha[k-1]=-one
    gamma=snp.zeros(k)
    gamma[0]=one
    beta[k-1]=one
    betaj=snp.zeros(k+1)
    for j in range(1,k):
        gamma[j]=one-sum(gamma[:j]/snp.arange(j+1,1,-1))
        for i in range(0,j+1):
            betaj[k-i-1]=(-one)**i*sympy.combinatorial.factorials.binomial(j,i)*gamma[j]
        beta=beta+betaj
    name=str(k)+'-step Adams-Bashforth method'
    return LinearMultistepMethod(alpha,beta,name=name)

def Adams_Moulton(k):
    r""" 
        Construct the k-step, Adams-Moulton method.
        The methods are implicit and have order k+1.
        They have the form:

        `y_{n+1} = y_n + h \sum_{j=0}^{k} \beta_j f(y_n-k+j+1)`

        They are generated using equation (1.9) and the equation in 
        Exercise 3 from Hairer & Wanner III.1, along with the binomial 
        expansion.

        **Examples**::

            >>> import linear_multistep_method as lm
            >>> am3=lm.Adams_Moulton(3)
            >>> am3.order()
            4

        References:
            [hairer1993]_
    """
    import sympy

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[k]=1
    alpha[k-1]=-1
    gamma=snp.zeros(k+1)
    gamma[0]=1
    beta[k]=1
    betaj=snp.zeros(k+1)
    for j in range(1,k+1):
        gamma[j]= -sum(gamma[:j]/snp.arange(j+1,1,-1))
        for i in range(0,j+1):
            betaj[k-i]=(-1)**i*sympy.combinatorial.factorials.binomial(j,i)*gamma[j]
            #betaj[k-i]=(-1.)**i*comb(j,i)*gamma[j]
        beta=beta+betaj
    name=str(k)+'-step Adams-Moulton method'
    return LinearMultistepMethod(alpha,beta,name=name)

def backward_difference_formula(k):
    r""" 
        Construct the k-step backward differentiation method.
        The methods are implicit and have order k.
        They have the form:

        `\sum_{j=0}^{k} \alpha_j y_{n+k-j+1} = h \beta_j f(y_{n+1})`

        They are generated using equation (1.22') from Hairer & Wanner III.1,
            along with the binomial expansion.

        **Examples**::

            >>> import linear_multistep_method as lm
            >>> bdf4=lm.backward_difference_formula(4)
            >>> bdf4.A_alpha_stability()
            146

        **References**:
            #.[hairer1993]_ pp. 364-365
    """
    import sympy

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    beta[k]=1
    gamma=snp.zeros(k+1)
    gamma[0]=1
    alphaj=snp.zeros(k+1)
    for j in range(1,k+1):
        gamma[j]= sympy.Rational(1,j)
        for i in range(0,j+1):
            alphaj[k-i]=(-1)**i*sympy.combinatorial.factorials.binomial(j,i)*gamma[j]
        alpha=alpha+alphaj
    name=str(k)+'-step BDF method'
    return LinearMultistepMethod(alpha,beta,name=name)

def elm_ssp2(k):
    r"""
    Returns the optimal SSP k-step linear multistep method of order 2.

    **Examples**::

        >>> import linear_multistep_method as lm
        >>> lm10=lm.elm_ssp2(10)
        >>> lm10.ssp_coefficient()
        8/9
    """
    import sympy

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[-1]=sympy.Rational(1,1)
    alpha[0]=sympy.Rational(-1,(k-1)**2)
    alpha[k-1]=sympy.Rational(-(k-1)**2+1,(k-1)**2)
    beta[k-1]=sympy.Rational(k,k-1)
    name='Optimal '+str(k)+'-step, 2nd order SSP method.'
    return LinearMultistepMethod(alpha,beta,name=name)

def sand_cc(s):
    r""" Construct Sand's circle-contractive method of order `p=2(s+1)`.

    **Examples**::

        >>> import linear_multistep_method as lm
        >>> cc4 = lm.sand_cc(4)
        >>> cc4.order()
        10
        >>> cc4.ssp_coefficient()
        1/8

    **References**:
        #. [sand1986]_
    """
    import sympy

    one  = sympy.Rational(1)
    zero = sympy.Rational(0)

    k = 2**s + 1
    p = 2*(s+1)

    Jn = [k,k-1]
    for i in range(1,s+1):
        Jn.append(k-1-2**i)

    alpha = snp.zeros(k+1)
    beta  = snp.zeros(k+1)

    # This is inefficient
    for j in Jn:
        tau_product = one
        tau_sum = zero
        tau = [one/(j-i) for i in Jn if i!=j]
        tau_product = np.prod(tau)
        tau_sum = np.sum(tau)
        beta[j] = tau_product**2
        alpha[j] = 2*beta[j]*tau_sum
    return LinearMultistepMethod(alpha,beta,'Sand circle-contractive')
        

def loadLMM(which='All'):
    """ 
    Load a set of standard linear multistep methods for testing.

    **Examples**::

        >>> from nodepy import lm
        >>> ebdf5 = lm.loadLMM('eBDF5')
        >>> ebdf5.is_zero_stable()
        True
    """
    import snp
    import sympy

    LM={}
    #================================================
    alpha=snp.array([-12,75,-200,300,-300,137])/sympy.Rational(137,1)
    beta=snp.array([60,-300,600,-600,300,0])/sympy.Rational(137,1)
    LM['eBDF5']=LinearMultistepMethod(alpha,beta,'eBDF 5')
    #================================================
    if which=='All':
        return LM
    else:
        return LM[which]


if __name__ == "__main__":
    import doctest
    doctest.testmod()



def collect_powers(s,v):
    import re
    m=s.count(v)
    for i in range(m,0,-1):
        pattern=re.compile(v+'(\*'+v+'){'+str(i)+'}')
        s=pattern.sub(v+'**'+str(i+1),s)
    return s

def collect_sums(s,v):
    import re
    m=s.count(v)
    for i in range(m,1,-1):
        pattern=re.compile('(\+'+v+'){'+str(i)+'}')
        s=pattern.sub(str(i)+'*'+v,s)
    return s

def mysimp(s):
    from sympy import sympify
    s=s.replace('1*','')
    s=s.replace('*1','')
    s=s.replace('--','')
    s=s.replace('+-','-')
    s=collect_powers(s,'c')
#    s=collect_sums(s,'c')
#    return str(sympify(s))
    return s


def addper(s):
    return s+'.'

#=====================================================
def get_substring(st,startpos):
#=====================================================
    """
        Extracts everything between an open and close
        brace from a string.

        INPUT:
            st       -- any string (usually a RootedTree)
            startpos -- an integer such that st[startpos] is
                        the open brace ('{') of the desired substring

        OUTPUT:
            A string containing everything from st[startpos] to
            the corresponding close brace ('}') (inclusive).
    """
    return st[startpos:open_to_close(st,startpos)+1]

#=====================================================
def open_to_close(st,startpos):
#=====================================================
    """ 
        Finds end of a substring enclosed by braces starting at startpos.
        Used by get_substring.
    """

    pos=startpos
    openchar=st[pos]
    if openchar=='{':  closechar='}'
    count=1
    while count>0:
        pos+=1
        if st[pos]==openchar:  count+=1
        if st[pos]==closechar: count-=1
    return pos
